unit uGlobalWorker;

interface

uses
  { System }
  System.Classes,
  System.Generics.Collections,
  { Uniform }
  uSettings;

type
  IGlobalWorker = interface
    ['{1F881750-D90A-4D4B-A7CB-BE7FCC92C020}']
    procedure ExecuteReport(ANumber: Integer);
    procedure Finalize;
    procedure Initialize(ASettings: TSettings; AThreadNumber: Integer);
  end;

  TGlobalWorker = class(TInterfacedObject, IGlobalWorker)
  strict protected
    FCurrentNumber: Integer;
    FReportStreams: TObjectList<TMemoryStream>;
    FSettings: TSettings;
    FThreadNumber: Integer;
    procedure AddToMergeList(AStream: TMemoryStream);
    function ConvertToPdf(AStream: TMemoryStream): TMemoryStream;
    procedure ExecuteReport(ANumber: Integer); virtual;
    procedure Finalize;
    function ProcessData: Boolean; virtual;
    procedure ProcessResult(AChart: Integer; AStream: TMemoryStream);
    function MergeReports: TMemoryStream;
    procedure SaveStream(AStream: TMemoryStream; AChart: Integer);
  public
    procedure Initialize(ASettings: TSettings; AThreadNumber: Integer);
  end;

implementation

uses
  { ReportBuilder }
  ppPDFDevice,
  ppArchiv,
  { System }
  System.SysUtils,
  { Uniform }
  uLogManager;

{ TGlobalWorker }

procedure TGlobalWorker.AddToMergeList(AStream: TMemoryStream);
var
  LReport: TMemoryStream;
begin
  LReport := TMemoryStream.Create;
  LReport.LoadFromStream(AStream);
  FReportStreams.Add(LReport);
end;

function TGlobalWorker.ConvertToPdf(AStream: TMemoryStream): TMemoryStream;
var
  LPDFDevice: TppPDFDevice;
  LReader: TppCustomArchiveReader;
begin
  LPDFDevice := TppPDFDevice.Create(nil);
  LReader := TppCustomArchiveReader.Create(nil);
  try
    LPDFDevice.PDFSettings.ScaleImages := False;
    LPDFDevice.StartPrintJob := True;
    Result := TMemoryStream.Create;
    LPDFDevice.OutputStream := Result;
    LReader.ArchiveStream := AStream;
    LPDFDevice.EndPrintJob := True;
    LPDFDevice.Publisher := LReader.Publisher;
    LReader.PrintToDevices;
    LPDFDevice.StartPrintJob := False;
  finally
    LPDFDevice.Free;
    LReader.Free;
  end;
end;

procedure TGlobalWorker.ExecuteReport(ANumber: Integer);
begin
  try
    FCurrentNumber := ANumber;
    ProcessData;
  except
    on E: Exception do
    begin
      LogManager.LogException('TGlobalWorker.ExecuteReport(' + IntToStr(FCurrentNumber) + ')', E);
    end;
  end;
end;

procedure TGlobalWorker.Finalize;
var
  LMerge: TMemoryStream;
begin
  if FSettings.DoMergeReports and (FReportStreams.Count > 0) then
  begin
    LMerge := MergeReports;
    try
      if FSettings.DoLogReports then
        SaveStream(LMerge, -1);
    finally
      LMerge.Free;
    end;
  end;
  FReportStreams.Free;
end;

procedure TGlobalWorker.Initialize(ASettings: TSettings; AThreadNumber: Integer);
begin
  FSettings := ASettings;
  FThreadNumber := AThreadNumber;
  FReportStreams := TObjectList<TMemoryStream>.Create;
end;

function TGlobalWorker.MergeReports: TMemoryStream;
var
  i: Integer;
  LHasMorePages: Boolean;
  LPDFDevice: TppPDFDevice;
  LReader: TppCustomArchiveReader;
begin
  LPDFDevice := TppPDFDevice.Create(nil);
  LReader := TppCustomArchiveReader.Create(nil);
  try
    LPDFDevice.PDFSettings.ScaleImages := False;
    LPDFDevice.StartPrintJob := True;
    Result := TMemoryStream.Create;
    LPDFDevice.OutputStream := Result;
    for i := 0 to FReportStreams.Count - 1 do
    begin
      LHasMorePages := (i < FReportStreams.Count - 1);
      LReader.ArchiveStream := FReportStreams[i];
      LPDFDevice.EndPrintJob := not LHasMorePages;
      LPDFDevice.Publisher := LReader.Publisher;
      LReader.PrintToDevices;
      if LHasMorePages then
      begin
        LPDFDevice.Reset;
        LReader.Reset;
      end;
      LPDFDevice.StartPrintJob := False;
    end;
  finally
    LPDFDevice.Free;
    LReader.Free;
  end;
end;

function TGlobalWorker.ProcessData: Boolean;
begin
  Result := False;
end;

procedure TGlobalWorker.ProcessResult(AChart: Integer; AStream: TMemoryStream);
var
  LResult: TMemoryStream;
begin
  try
    LResult := nil;
    if FSettings.DoPrintDirectlyToPdf then
    begin
      LResult := TMemoryStream.Create;
      LResult.LoadFromStream(AStream);
    end
    else
    begin
      if FSettings.DoMergeReports then
        AddToMergeList(AStream)
      else
        LResult := ConvertToPdf(AStream);
    end;
    try
      if FSettings.DoLogReports then
        SaveStream(LResult, AChart);
    finally
      LResult.Free;
    end;
  finally
    AStream.Free;
  end;
end;

procedure TGlobalWorker.SaveStream(AStream: TMemoryStream; AChart: Integer);
var
  LEventDateStr: string;
  LLogFolder: string;
  LNow: TDateTime;
  LReportFile: string;
  LSubFolder: string;
begin
  if Assigned(AStream) then
  begin
    LReportFile := '';
    LNow := Now;
    LSubFolder := FormatDateTime('yyyymmdd', LNow);
    LEventDateStr := FormatDateTime('hhnnss', LNow);
    LLogFolder := IncludeTrailingPathDelimiter(FSettings.LogFolder) + LSubFolder;
    ForceDirectories(LLogFolder);
    if (AChart > 0) then
      LReportFile := IncludeTrailingPathDelimiter(LLogFolder) + LEventDateStr + '_' + IntToStr(FThreadNumber) + '_' + IntToStr(AChart) + '_' + IntToStr(FCurrentNumber) + '.pdf'
    else
      LReportFile := IncludeTrailingPathDelimiter(LLogFolder) + LEventDateStr + '_' + IntToStr(FThreadNumber) + '.pdf';
    AStream.SaveToFile(LReportFile);
  end;
end;

end.
