unit uLogManager;

interface

uses
  { System }
  System.Classes,
  System.SyncObjs,
  System.SysUtils,
  { Uniform }
  uLogging,
  uSettings;

const
  LOG_FILE_PREFIX = 'ReportBuilder_Test';

type
  ILogManager = interface
    ['{1541E8C1-A74C-4741-A799-CDC22318E3AA}']
    function GetIsDestroying: Boolean;
    procedure Initialize(const ALogFolder: string);
    procedure LogException(const APrefix: string; const E: Exception);
    procedure LogMessage(const ALogType: TLogMessageType; const AMessage: string);
    procedure LogMessageError(const AMessage: string);
    procedure LogMessageInfo(const AMessage: string);
    procedure LogMessageWarning(const AMessage: string);
    procedure LogSettings(const ASettings: TSettings);
    procedure MemoLog(AMemoLogLines: TStrings);
    procedure SetIsDestroying(Value: Boolean);
    procedure SetLogMessageEvent(AOnLogMessage: TOnLogMessage);
    property IsDestroying: Boolean read GetIsDestroying write SetIsDestroying;
  end;

function LogManager: ILogManager;

implementation

{$IFDEF DEBUG}
//uses
//  madExcept,
//  madStackTrace;
{$ENDIF}

type
  TLogManager = class(TInterfacedObject, ILogManager)
  strict private
    class var FCriticalSection: TCriticalSection;
    class var FLogManager: ILogManager;
  strict private
    FIsDestroying: Boolean;
    FIsStarted: Boolean;
    FLogMessageThread: TLogMessageThread;
    { IServiceLogManager }
    function GetIsDestroying: Boolean;
    procedure Initialize(const ALogFolder: string);
    procedure LogException(const APrefix: string; const E: Exception);
    procedure LogMessage(const ALogType: TLogMessageType; const AMessage: string);
    procedure LogMessageError(const AMessage: string);
    procedure LogMessageInfo(const AMessage: string);
    procedure LogMessageWarning(const AMessage: string);
    procedure LogSettings(const ASettings: TSettings);
    procedure MemoLog(AMemoLogLines: TStrings);
    procedure SetIsDestroying(Value: Boolean);
    procedure SetLogMessageEvent(AOnLogMessage: TOnLogMessage);
  public
    class function Instance: ILogManager; static;
    class constructor CreateClass;
    class destructor DestroyClass;
  public
    constructor Create;
    destructor Destroy; override;
  end;

function LogManager: ILogManager;
begin
  Result := TLogManager.Instance;
end;

{ TLogManager }

constructor TLogManager.Create;
begin
  inherited Create;
  FIsStarted := False;
  FLogMessageThread := TLogMessageThread.Create;
end;

class constructor TLogManager.CreateClass;
begin
  FCriticalSection := TCriticalSection.Create;
end;

destructor TLogManager.Destroy;
begin
  FLogMessageThread.Terminate;
  FLogMessageThread.WaitFor;
  FLogMessageThread.Free;
  FLogMessageThread := nil;
  inherited Destroy;
end;

class destructor TLogManager.DestroyClass;
begin
  FCriticalSection.Free;
end;

function TLogManager.GetIsDestroying: Boolean;
begin
  Result := FIsDestroying;
end;

procedure TLogManager.Initialize(const ALogFolder: string);
var
  LLogFile: string;
begin
  LLogFile := IncludeTrailingPathDelimiter(ALogFolder) + LOG_FILE_PREFIX + FormatDateTime('_yyyymmdd', Now) + '.log';
  ForceDirectories(ALogFolder);
  if not FIsStarted then
  begin
    FLogMessageThread.Initialize(LLogFile);
    FLogMessageThread.Start;
    FIsStarted := True;
  end;
end;

class function TLogManager.Instance: ILogManager;
begin
  if not Assigned(FLogManager) then
  begin
    FCriticalSection.Enter;
    try
      if not Assigned(FLogManager) then
      begin
        FLogManager := TLogManager.Create;
      end;
    finally
      FCriticalSection.Leave;
    end;
  end;
  Result := FLogManager;
end;

procedure TLogManager.LogException(const APrefix: string; const E: Exception);
begin
  {.$IFDEF DEBUG}
   // LogMessage(lmtError, APrefix + ': StackTrace =' + sLineBreak + string(StackTrace));
   // var LBugReport := NewException(etNormal, E).GetBugReport(True);
   // LogMessage(lmtError, APrefix + sLineBreak + LBugReport);
  {.$ELSE}
    LogMessage(lmtError, APrefix + sLineBreak + E.message);
  {.$ENDIF}
end;

procedure TLogManager.LogMessage(const ALogType: TLogMessageType;
  const AMessage: string);
var
  LLogMsg: TLogMsgData;
begin
  if GetIsDestroying then
  begin
    Exit;
  end;
  LLogMsg.MsgType := ALogType;
  LLogMsg.Msg := AMessage;
  LLogMsg.ThreadId := TThread.CurrentThread.ThreadID;
  FLogMessageThread.LogMessageQueue.Enqueue(LLogMsg);
end;

procedure TLogManager.LogMessageError(const AMessage: string);
begin
  LogMessage(lmtError, AMessage);
end;

procedure TLogManager.LogMessageInfo(const AMessage: string);
begin
  LogMessage(lmtInfo, AMessage);
end;

procedure TLogManager.LogMessageWarning(const AMessage: string);
begin
  LogMessage(lmtWarning, AMessage);
end;

procedure TLogManager.LogSettings(const ASettings: TSettings);
begin
  LogMessage(lmtInfo, ASettings.WriteSettings);
end;

procedure TLogManager.MemoLog(AMemoLogLines: TStrings);
begin
  FLogMessageThread.MemoLog(AMemoLogLines);
end;

procedure TLogManager.SetIsDestroying(Value: Boolean);
begin
  FIsDestroying := Value;
  FLogMessageThread.Terminate;
end;

procedure TLogManager.SetLogMessageEvent(AOnLogMessage: TOnLogMessage);
begin
  FLogMessageThread.SetLogMessageEvent(AOnLogMessage);
end;

end.
