unit uLogging;

interface

uses
  { System }
  System.Classes,
  { Uniform }
  uThreadSafe;

type
  TLogMessageType = (lmtInfo, lmtWarning, lmtError);

  TLogMessageTypeHelper = record helper for TLogMessageType
  public
    function Prefix: string;
  end;

  TLogMsgData = packed record
    Msg: string;
    MsgType: TLogMessageType;
    ThreadId: Cardinal;
  end;

  TOnLogMessage = procedure (const ALogType: TLogMessageType; const AMessage: string) of object;

  TLogMessageQueue = class(TThreadSafeQueue<TLogMsgData>);

  TLogMessageThread = class(TThread)
  private
    FLogFileName: string;
    FLogMessageQueue: TLogMessageQueue;
    FLineCount: Integer;
    FMemoLogLines: TStrings;
    FMemoLogLinesBuffer: TStringList;
    FOnLogMessage: TOnLogMessage;
    procedure DoLogMessage(const ALogType: TLogMessageType; const AMessage: string);
  public
    property LogMessageQueue: TLogMessageQueue read FLogMessageQueue;
    constructor Create; reintroduce;
    destructor Destroy; override;
    procedure Execute; override;
    procedure Initialize(const ALogFileName: string);
    procedure MemoLog(AMemoLogLines: TStrings);
    procedure SetLogMessageEvent(AOnLogMessage: TOnLogMessage);
  end;

implementation

uses
  { System }
  System.IOUtils,
  System.SysUtils,
  System.Types,
  Vcl.Forms;

const
  DEQUEUE_TIMEOUT = 500;

{ TLogMessageTypeHelper }

function TLogMessageTypeHelper.Prefix: string;
begin
  case Self of
    lmtWarning: Result := '[WARN ] ';
    lmtError  : Result := '[ERROR] ';
  else
    Result := '[INFO ] ';
  end;
end;

{ TLogMessageThread }

constructor TLogMessageThread.Create;
begin
  inherited Create(True);
  FMemoLogLinesBuffer := TStringList.Create;
  FLogMessageQueue := TLogmessageQueue.Create(DEQUEUE_TIMEOUT);
end;

destructor TLogMessageThread.Destroy;
begin
  FMemoLogLines := nil;
  FMemoLogLinesBuffer.Free;
  FLogMessageQueue.Free;
  inherited Destroy;
end;

procedure TLogMessageThread.DoLogMessage(const ALogType: TLogMessageType;
  const AMessage: string);
var
  LEventDateStr: string;
  LStream: TFileStream;
  LLogWriter: TStreamWriter;
  LText: string;
begin
  LEventDateStr := FormatDateTime('[yyyy.mm.dd hh:nn:ss] ', Now);
  LText := LEventDateStr + ALogType.Prefix + AMessage;
  if Assigned(FMemoLogLines) then
  begin
    TThread.Queue(nil,
      procedure
      begin
        while FMemoLogLines.Count > 10000 do
        begin
          FMemoLogLines.Delete(0);
        end;
        FMemoLogLines.Add(LText);
      end);
  end
  else if Assigned(Application.MainForm) then
  begin
    FMemoLogLinesBuffer.Add(LText);
  end;
  if Assigned(FOnLogMessage) then
  begin
    FOnLogMessage(ALogType, AMessage);
  end;
  if FLogFileName = '' then
  begin
    Writeln(System.Output, LText);
  end
  else
  begin
    if not FileExists(FLogFileName)then
    begin
      TDirectory.CreateDirectory(ExtractFilePath(FLogFileName));
      LStream := TFileStream.Create(FLogFileName, fmCreate);
    end
    else
    begin
      LStream := TFileStream.Create(FLogFileName, fmOpenWrite or fmShareDenyNone);
      LStream.Seek(0, soEnd);
    end;
    try
      LLogWriter := TStreamWriter.Create(LStream);
      try
        LLogWriter.WriteLine(LText);
        LLogWriter.Flush
      finally
        LLogWriter.Free();
      end;
    finally
      LStream.Free;
    end;
    if FLineCount < 100 then
    begin
      Inc(FLineCount)
    end
    else
    begin
      FLineCount := 0;
    end;
  end;
end;

procedure TLogMessageThread.Execute;
var
  LMessage: TLogMsgData;
begin
  while not Terminated do begin
    Sleep(1);
    if FLogMessageQueue.Dequeue(LMessage) = TWaitResult.wrSignaled then
    begin
      DoLogMessage(LMessage.MsgType, Format('[%d]>%s', [LMessage.ThreadId, LMessage.Msg]));
    end;
  end;
  FLogMessageQueue.DoShutDown;
end;

procedure TLogMessageThread.Initialize(const ALogFileName: string);
begin
  FLogFileName := ALogFileName;
  FLineCount := 0;
end;

procedure TLogMessageThread.MemoLog(AMemoLogLines: TStrings);
begin
  FMemoLogLines := AMemoLogLines;
  if Assigned(FMemoLogLines) then
  begin
    FMemoLogLines.AddStrings(FMemoLogLinesBuffer);
    FMemoLogLinesBuffer.Clear;
  end;
end;

procedure TLogMessageThread.SetLogMessageEvent(AOnLogMessage: TOnLogMessage);
begin
  FOnLogMessage := AOnLogMessage;
end;

end.
