unit uThreadSafe;

interface

uses
  { System }
  System.Generics.Collections,
  System.SyncObjs;

type
  TThreadSafeQueue<T> = class
  private
    FCriticalSection: TCriticalSection;
    FDequeueTimeout: UInt64;
    FEvent: TEvent;
    FQueue: TQueue<T>;
    FShutDown: Boolean;
    procedure Clear;
  protected
    procedure ClearItem(AItem: T); virtual;
  public
    function Dequeue(out Item: T): TWaitResult;
    function Enqueue(const Item: T): Boolean;
    function QueueSize: integer;
  public
    constructor Create(const ADequeueTimeout: UInt64); reintroduce; virtual;
    destructor Destroy; override;
    procedure DoShutDown;
  end;

implementation

uses
  System.SysUtils;

{ TThreadSafeQueue<T> }

procedure TThreadSafeQueue<T>.Clear;
begin
  while FQueue.Count > 0 do
  begin
    ClearItem(FQueue.Dequeue);
  end;
end;

procedure TThreadSafeQueue<T>.ClearItem(AItem: T);
begin
end;

constructor TThreadSafeQueue<T>.Create(const ADequeueTimeout: UInt64);
begin
  inherited Create;
  FShutDown := False;
  FQueue := TQueue<T>.Create;
  FCriticalSection := TCriticalSection.Create;
  FEvent := TEvent.Create(nil, True, False, '');
  FDequeueTimeout := ADequeueTimeout;
end;

function TThreadSafeQueue<T>.Dequeue(out Item: T): TWaitResult;
var
  LWaitResult: TWaitResult;
begin
  if FShutDown then
    Exit(TWaitResult.wrAbandoned);
  LWaitResult := FEvent.WaitFor(FDequeueTimeout);
  if LWaitResult = TWaitResult.wrSignaled then
  begin
    if FShutDown then
      Exit(TWaitResult.wrAbandoned);
    FCriticalSection.Enter;
    try
      if FQueue.Count = 0 then
      begin
        FEvent.ResetEvent;
        Exit(TWaitResult.wrTimeout);
      end;
      Item := FQueue.Extract;
      Result := TWaitResult.wrSignaled;
    finally
      FCriticalSection.Leave;
    end;
  end
  else
  begin
    Result := lWaitResult;
  end;
end;

destructor TThreadSafeQueue<T>.Destroy;
begin
  DoShutDown;
  FQueue.Free;
  FCriticalSection.Free;
  FEvent.Free;
  inherited Destroy;
end;

procedure TThreadSafeQueue<T>.DoShutDown;
begin
  FShutDown := True;
  FCriticalSection.Enter;
  try
    Clear;
  finally
    FCriticalSection.Leave;
  end;
end;

function TThreadSafeQueue<T>.Enqueue(const Item: T): Boolean;
const
  cRetryCount: Byte = 5;
var
  LCount: Integer;
begin
  if FShutDown then
    Exit(False);
  Result := False;
  LCount := 0;
  while LCount < cRetryCount do
  begin
    if LCount > 0 then
    begin
      Sleep(LCount * 10);
    end;
    FCriticalSection.Enter;
    try
      if FQueue.Count >= 100 then
      begin
        Inc(LCount);
        Continue;
      end;
      FQueue.Enqueue(Item);
      Result := True;
      FEvent.SetEvent;
      Break;
    finally
      FCriticalSection.Leave;
    end;
  end;
end;

function TThreadSafeQueue<T>.QueueSize: integer;
begin
  FCriticalSection.Enter;
  try
    Result := FQueue.Count;
  finally
    FCriticalSection.Leave;
  end;
end;

end.
