unit uWorkerManager;

interface

uses
  { System }
  System.Generics.Collections,
  System.SysUtils,
  System.SyncObjs,
  { Uniform }
  uSettings,
  uThread;

type
  TWorkerManager = class(TObject)
  private
    FCriticalSection: TCriticalSection;
    FThreadList: TList<TWorkerThread>;
    FFinishedProc: TProc;
    procedure ThreadOnTerminate(Sender: TObject);
    procedure Finished;
  protected
    procedure OnTerminateEvent(Sender:TObject);
  public
    constructor Create;
    destructor Destroy; override;
    function CountRunningThreads: Integer;
    procedure Execute(ASettings: TSettings; AFinishedProc: TProc);
    procedure Stop;
  end;

implementation

uses
  { Uniform }
  uLogManager,
  uWorker;

{ TWorkerManager }

function TWorkerManager.CountRunningThreads: Integer;
begin
  Result := 0;
  if Assigned(FThreadList) then
  begin
    Result := FThreadList.Count;
  end;
end;

constructor TWorkerManager.Create;
begin
  inherited Create;
  FCriticalSection := TCriticalSection.Create;
  FThreadList := TList<TWorkerThread>.Create;
end;

destructor TWorkerManager.Destroy;
begin
  FCriticalSection.Destroy;
  FThreadList.Free;
  inherited Destroy;
end;

procedure TWorkerManager.Execute(ASettings: TSettings; AFinishedProc: TProc);
var
  LThreadCount: Integer;
  LWorker: TChartRandom_Worker;
begin
  LogManager.LogSettings(ASettings);
  FFinishedProc := AFinishedProc;
  for LThreadCount := 1 to ASettings.NumberOfThreads do
  begin
    var LThread := TWorkerThread.Create(ASettings, LThreadCount);
    LWorker := TChartRandom_Worker.Create;
    LWorker.Initialize(ASettings, LThread.Number);
    LThread.Worker := LWorker;
    LThread.OnTerminate := ThreadOnTerminate;
    LThread.FreeOnTerminate := True;
    FCriticalSection.Enter;
    try
      FThreadList.Add(LThread);
    finally
      FCriticalSection.Leave;
    end;
  end;

  for var LThread in FThreadList do
  begin
    LThread.Start;
  end;
end;

procedure TWorkerManager.Finished;
begin
  if Assigned(FFinishedProc) then
  begin
    FFinishedProc;
  end;
end;

procedure TWorkerManager.OnTerminateEvent(Sender: TObject);
var
  LThread: TWorkerThread;
begin
  if (Sender is TWorkerThread) then
  begin
    LThread := TWorkerThread(Sender);
    LThread.Finalize;
    LogManager.LogMessageInfo(Format('Terminate Thread %d', [LThread.Number]));
  end;
end;

procedure TWorkerManager.Stop;
var
  LThread: TWorkerThread;
begin
  LogManager.LogMessageInfo('Stop...');
  for LThread in FThreadList do
  begin
    LThread.Terminate;
  end;
end;

procedure TWorkerManager.ThreadOnTerminate(Sender: TObject);
var
  LTerminatedThread: TWorkerThread;
begin
  if Sender is TWorkerThread then
  begin
    FCriticalSection.Enter;
    try
      LTerminatedThread := TWorkerThread(Sender);
      OnTerminateEvent(LTerminatedThread);
      FThreadList.Remove(LTerminatedThread);

      if FThreadList.Count = 0 then
      begin
        Finished;
      end;
    finally
      FCriticalSection.Leave;
    end;
  end;
end;

end.
